import shutil
from textwrap import dedent

import importlib_resources

from vasp_manager import VaspManager


def test_vmg_in_order(tmp_path):
    original_calculations_dir = (
        importlib_resources.files("vasp_manager") / "tests" / "calculations"
    )
    temp_calculations_dir = tmp_path / "calculations"
    shutil.copytree(
        original_calculations_dir,
        temp_calculations_dir,
        dirs_exist_ok=True,
        symlinks=True,
        ignore=shutil.ignore_patterns(
            "rlx*",
            "static",
            "bulkmod",
            "elastic",
            "material_hit_errors",
            "material_needs_archive",
            "material_needs_rerun",
        ),
    )

    calculation_types = [
        "rlx-coarse",
        "rlx",
        "static",
        "bulkmod",
        "elastic",
    ]
    material_dirs = [
        p for p in sorted(list(temp_calculations_dir.glob("*"))) if p.is_dir()
    ]

    for i, calculation_type in enumerate(calculation_types):
        calculation_type_subset = calculation_types[: i + 1]
        vmg = VaspManager(
            calculation_types=calculation_type_subset,
            material_dirs=material_dirs,
            use_multiprocessing=False,
            to_rerun=True,
            to_submit=True,
        )
        results = vmg.run_calculations()
        if i != 0:
            previous_calc_types = calculation_type_subset[:i]
            # check results
            for material in results:
                for pc in previous_calc_types:
                    match pc:
                        case "rlx-coarse":
                            assert results[material][pc] == "done"
                        case "rlx":
                            assert isinstance(results[material][pc]["total_dV"], float)
                        case "static":
                            assert isinstance(
                                results[material][pc]["final_energy"], float
                            )
                        case "bulkmod":
                            assert isinstance(results[material][pc]["B"], float)
                        case _:
                            raise Exception

            # check summary
            summary = vmg.summary(as_string=False)
            for pc in previous_calc_types:
                assert summary["n_total"] == summary[pc]["n_finished"]

        for p in material_dirs:
            mat_name = p.name
            orig_mode_dir = original_calculations_dir / mat_name / calculation_type
            temp_mode_dir = p / calculation_type
            shutil.copytree(
                orig_mode_dir,
                temp_mode_dir,
                dirs_exist_ok=True,
                symlinks=True,
                ignore=shutil.ignore_patterns("elastic_constants.txt"),
            )

        # test string summary when unfinished
        if i == (len(calculation_types) - 1):
            summary = vmg.summary(as_string=True, print_unfinished=True)
            correct_summary = dedent(
                """\
            Total Materials = 2
            ------------------------------
            RLX-COARSE  2/2 completed
            RLX         2/2 completed
            STATIC      2/2 completed
            BULKMOD     2/2 completed
            ELASTIC     0/2 completed
                        2 not completed
            Unfinished ELASTIC: ['material', 'material_spinu']
            """
            )
            assert summary == correct_summary

    vmg = VaspManager(
        calculation_types=calculation_types,
        material_dirs=material_dirs,
        use_multiprocessing=False,
        to_rerun=True,
        to_submit=True,
    )
    results = vmg.run_calculations()
    for material in results:
        assert results[material]["rlx-coarse"] == "done"
        assert isinstance(results[material]["rlx"]["initial_spacegroup"], int)
        assert isinstance(results[material]["rlx"]["relaxed_spacegroup"], int)
        assert isinstance(results[material]["rlx"]["total_dV"], float)
        assert isinstance(results[material]["static"]["final_energy"], float)
        assert isinstance(results[material]["bulkmod"]["B"], float)
        assert isinstance(results[material]["elastic"]["B_VRH"], float)

    summary = vmg.summary(as_string=False)
    for calculation_type in calculation_types:
        assert summary["n_total"] == summary[calculation_type]["n_finished"]


def test_vmg_with_skipping(tmp_path):
    original_calculations_dir = (
        importlib_resources.files("vasp_manager") / "tests" / "calculations"
    )
    temp_calculations_dir = tmp_path / "calculations"
    # skip static and bulkmod when copying
    shutil.copytree(
        original_calculations_dir,
        temp_calculations_dir,
        dirs_exist_ok=True,
        symlinks=True,
        ignore=shutil.ignore_patterns(
            "static",
            "bulkmod",
            "material_hit_errors",
            "material_needs_archive",
            "material_needs_rerun",
        ),
    )

    calculation_types = [
        "rlx-coarse",
        "rlx",
        "static",
        "bulkmod",
        "elastic",
    ]
    material_dirs = [
        p for p in sorted(list(temp_calculations_dir.glob("*"))) if p.is_dir()
    ]

    vmg = VaspManager(
        calculation_types=calculation_types,
        material_dirs=material_dirs,
        use_multiprocessing=True,
        to_rerun=True,
        to_submit=True,
    )
    results = vmg.run_calculations()
    for material in results:
        for calc_type in calculation_types:
            match calc_type:
                case "rlx-coarse":
                    assert results[material][calc_type] == "done"
                case "rlx":
                    assert isinstance(results[material][calc_type]["total_dV"], float)
                case "static" | "bulkmod":
                    assert results[material][calc_type] is None
                case "elastic":
                    assert isinstance(results[material][calc_type]["B_VRH"], float)
                case _:
                    raise Exception

    summary = vmg.summary(as_string=False)
    for pc in calculation_types:
        match pc:
            case "rlx-coarse" | "rlx" | "elastic":
                assert summary["n_total"] == summary[pc]["n_finished"]
            case "bulkmod" | "static":
                assert summary[pc]["n_finished"] == 0
            case _:
                raise Exception



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from pymatgen.analysis.eos import BirchMurnaghan
from pymatgen.core import Structure
from pymatgen.io.vasp import Vasprun

if TYPE_CHECKING:
    from vasp_manager.types import WorkingDirectory

logger = logging.getLogger(__name__)


class BulkmodAnalyzer:
    def __init__(
        self,
        calc_dir: WorkingDirectory,
        rounding_precision: int = 3,
    ) -> None:
        """
        Args:
            calc_dir: path to bulkmod calculation directory
            rounding_precision: precision to round calculated quantities
        """
        self.calc_dir = Path(calc_dir)
        self.rounding_precision = rounding_precision
        self._results: dict

    @property
    def calc_dir(self) -> Path:
        return self._calc_dir

    @calc_dir.setter
    def calc_dir(self, value: Path) -> None:
        if not value.exists():
            raise ValueError(f"Could not set calc_dir to {value} as it does not exist")
        self._calc_dir = value

    @property
    def rounding_precision(self) -> int:
        return self._rounding_precision

    @rounding_precision.setter
    def rounding_precision(self, value: int) -> None:
        if not isinstance(value, int):
            raise ValueError(
                f"Could not set rounding_precision to {value}, rounding precision must"
                " be an int"
            )
        self._rounding_precision = value

    @staticmethod
    def analyze_bulkmod(calc_dir: Path, rounding_precision: int) -> dict:
        """
        Fit an EOS to calculate the bulk modulus from a finished bulkmod calculation
        """
        strain_dirs = [path for path in calc_dir.glob("strain*") if path.is_dir()]
        strain_dirs = sorted(strain_dirs, key=lambda d: int(d.name.split("_")[-1]))
        volumes = []
        final_energies = []
        for i, strain_dir in enumerate(strain_dirs):
            poscar_path = strain_dir / "POSCAR"
            # search for vasprun.xml or vasprun.xml.gz
            vasprun_glob = list(strain_dir.glob("vasprun.xml*"))
            if len(vasprun_glob) == 0:
                raise Exception(f"No vasprun.xml available at {strain_dir}")
            vasprun_path = vasprun_glob[0]
            volume = Structure.from_file(poscar_path).volume
            vasprun = Vasprun(
                filename=vasprun_path,
                parse_dos=False,
                parse_eigen=False,
                parse_potcar_file=False,
            )
            final_energy = vasprun.final_energy
            volumes.append(volume)
            final_energies.append(final_energy)
        logger.debug(f"Volumes:\n\t{volumes}")
        logger.debug(f"Final Energies:\n\t{final_energies}")
        eos_analyzer = BirchMurnaghan(volumes, final_energies)
        eos_analyzer.fit()
        bulk_modulus = np.round(eos_analyzer.b0_GPa, rounding_precision)
        logger.debug(f"BULK MODULUS: {bulk_modulus}")
        b_dict = {"B": bulk_modulus}
        return b_dict

    @property
    def results(self) -> dict:
        if getattr(self, "_results", None) is None:
            self._results = self.analyze_bulkmod(self.calc_dir, self.rounding_precision)
        return self._results



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

from __future__ import annotations

from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import NDArray
from pymatgen.core import Structure

from vasp_manager.utils import pgrep

if TYPE_CHECKING:
    from vasp_manager.types import Floating, WorkingDirectory


class ElasticAnalyzer:
    def __init__(
        self,
        cij: NDArray,
        structure: Structure,
        rounding_precision: int = 3,
    ) -> None:
        """
        Args:
            cij: stiffness tensor in Voigt notation in units of GPa
            structure: the associated pymatgen structure
            rounding_precision: precision to round calculated quantities

        Note:
            Be careful with cij! VASP does not output stiffness tensors in
            typical Voigt notation. See change_elastic_constants_from_vasp
            method.
        """
        self.cij = cij
        self.structure = structure
        self.rounding_precision = rounding_precision
        self._results: dict

    @property
    def cij(self) -> NDArray:
        return self._cij

    @cij.setter
    def cij(self, values: NDArray) -> None:
        values = np.asarray(values)
        if values.shape != (6, 6):
            raise ValueError(
                "cij must be a 6x6 array," f"found shape={values.shape} instead"
            )
        if values.dtype != float:
            raise TypeError(
                "cij must be a 6x6 array of floats," f" found type={values.dtype} instead"
            )
        self._cij = values

    @property
    def structure(self) -> Structure:
        return self._structure

    @structure.setter
    def structure(self, value: Structure) -> None:
        if not isinstance(value, Structure):
            raise TypeError(
                "Structure must be pymatgen Structure object,"
                f" found type={type(value)} instead"
            )
        self._structure = value

    @property
    def density(self) -> float:
        return self.structure.density

    @property
    def rounding_precision(self) -> int:
        return self._rounding_precision

    @rounding_precision.setter
    def rounding_precision(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError(
                f"Could not set rounding_precision to {value}. rounding_precision must"
                " be an int, found type={type(value)} instead)"
            )
        self._rounding_precision = value

    @staticmethod
    def change_elastic_constants_from_vasp(vasp_elastic_tensor: NDArray) -> NDArray:
        """
        Args:
            vasp_elastic_tensor: elastic tensor from VASP OUTCAR

        Returns:
            elastic_tensor: reordered to match Voigt notation

        Note:
            VASP ordering of elastic constants is not the same as those expected by
            my equations below

            We should expect from Voigt notation:
            1: 11 or xx
            2: 22 or yy
            3: 33 or zz
            4: 23 or yz
            5: 13 or xz
            6: 12 or xy
            but VASP differs as it presents 1, 2, 3, 6 (xy), 4 (yz), 5 (xz)
            This function performs swapping to match expectations
        """
        elastic_tensor = np.asarray(vasp_elastic_tensor).copy()
        for j in range(6):
            elastic_tensor[3, j], elastic_tensor[4, j], elastic_tensor[5, j] = (
                elastic_tensor[4, j],
                elastic_tensor[5, j],
                elastic_tensor[3, j],
            )
        for i in range(6):
            elastic_tensor[i, 3], elastic_tensor[i, 4], elastic_tensor[i, 5] = (
                elastic_tensor[i, 4],
                elastic_tensor[i, 5],
                elastic_tensor[i, 3],
            )
        return elastic_tensor

    @staticmethod
    def get_compliance_tensor(cij: NDArray) -> NDArray:
        """
        Args:
            cij: stiffness tensor

        Returns:
            sij: compliance tensor
        """
        sij = np.linalg.inv(cij)
        return sij

    @staticmethod
    def get_VRH_average(mod1: Floating, mod2: Floating) -> Floating:
        """
        Args:
            mod1: B or G Voigt
            mod2: B or G Reuss

        Returns:
            VRH_average
        """
        return (mod1 + mod2) / 2.0

    @staticmethod
    def get_B_Reuss(sij: NDArray) -> float:
        """
        Args:
            sij: compliance tensor

        Returns:
            B_Reuss: Reuss bulk modulus
        """
        B_Reuss = 1 / (
            (sij[0, 0] + sij[1, 1] + sij[2, 2]) + 2 * (sij[0, 1] + sij[1, 2] + sij[2, 0])
        )
        return B_Reuss

    @staticmethod
    def get_B_Voigt(cij: NDArray) -> float:
        """
        Args:
            cij: stiffness tensor

        Returns:
            B_Voigt: Voigt bulk modulus
        """
        B_Voigt = (
            (cij[0, 0] + cij[1, 1] + cij[2, 2]) + 2 * (cij[0, 1] + cij[1, 2] + cij[2, 0])
        ) / 9
        return B_Voigt

    @staticmethod
    def get_G_Reuss(sij: NDArray) -> float:
        """
        Args:
            sij: compliance tensor

        Returns:
            G_Reuss: Reuss shear modulus
        """
        G_Reuss = 15 / (
            4 * (sij[0, 0] + sij[1, 1] + sij[2, 2])
            - 4 * (sij[0, 1] + sij[1, 2] + sij[2, 0])
            + 3 * (sij[3, 3] + sij[4, 4] + sij[5, 5])
        )
        return G_Reuss

    @staticmethod
    def get_G_Voigt(cij: NDArray) -> float:
        """
        Args:
            cij: stiffness tensor

        Returns:
            G_Voigt: Voigt shear modulus
        """
        G_Voigt = (
            (cij[0, 0] + cij[1, 1] + cij[2, 2])
            - (cij[0, 1] + cij[1, 2] + cij[2, 0])
            + 3 * (cij[3, 3] + cij[4, 4] + cij[5, 5])
        ) / 15
        return G_Voigt

    @staticmethod
    def get_vl(
        bulkmod: Floating,
        shearmod: Floating,
        density: Floating,
    ) -> Floating:
        """
        Returns:
            vl: longitudinal speed of sound
        """
        vl = np.sqrt((bulkmod + 4 / 3 * shearmod) / density)
        return vl

    @staticmethod
    def get_vt(
        shearmod: Floating,
        density: Floating,
    ) -> Floating:
        """
        Returns:
            vl: transverse speed of sound
        """
        vt = np.sqrt(shearmod / density)
        return vt

    @staticmethod
    def get_vs(
        bulkmod: Floating,
        shearmod: Floating,
        density: Floating,
    ) -> Floating:
        """
        Returns:
            vs: speed of sound
        """
        vl = np.sqrt((bulkmod + 4 / 3 * shearmod) / density)
        vt = np.sqrt(shearmod / density)
        vs = (1 / 3 * (1 / vl**3 + 2 / vt**3)) ** (-1 / 3)
        return vs

    @staticmethod
    def check_elastically_unstable(cij: NDArray) -> bool:
        """
        Returns:
            True if compound is elastically unstable
        """
        eigenvalues, eigenvectors = np.linalg.eig(cij)
        born_critera_satisfied = np.all(eigenvalues > 0)
        elastically_unstable = not born_critera_satisfied
        return elastically_unstable

    @cached_property
    def sij(self) -> NDArray:
        # DO NOT ROUND
        return self.get_compliance_tensor(self.cij)

    @cached_property
    def b_reuss(self) -> Floating:
        return np.round(self.get_B_Reuss(self.sij), self.rounding_precision)

    @cached_property
    def b_voigt(self) -> Floating:
        return np.round(self.get_B_Voigt(self.cij), self.rounding_precision)

    @cached_property
    def b_vrh(self) -> Floating:
        return np.round(
            self.get_VRH_average(self.b_reuss, self.b_voigt), self.rounding_precision
        )

    @cached_property
    def g_reuss(self) -> Floating:
        return np.round(self.get_G_Reuss(self.sij), self.rounding_precision)

    @cached_property
    def g_voigt(self) -> Floating:
        return np.round(self.get_G_Voigt(self.cij), self.rounding_precision)

    @cached_property
    def g_vrh(self) -> Floating:
        return np.round(
            self.get_VRH_average(self.g_reuss, self.g_voigt), self.rounding_precision
        )

    @cached_property
    def elastically_unstable(self) -> bool:
        return self.check_elastically_unstable(self.cij)

    @cached_property
    def vt(self) -> Floating:
        return np.round(self.get_vt(self.g_vrh, self.density), self.rounding_precision)

    @cached_property
    def vl(self) -> Floating:
        return np.round(
            self.get_vl(self.b_vrh, self.g_vrh, self.density), self.rounding_precision
        )

    @cached_property
    def vs(self) -> Floating:
        return np.round(
            self.get_vs(self.b_vrh, self.g_vrh, self.density), self.rounding_precision
        )

    def _analyze_elastic(self, properties: list | None = None) -> dict:
        """
        Grabs important quantities from the elastic calculation results

        Args:
            properties (list): names of properties to calculate

        Returns:
            elastic_dict (dict): dict of extracted info from
                elastic constants
        """
        properties_map = {
            "B_Reuss": "b_reuss",
            "B_Voigt": "b_voigt",
            "B_VRH": "b_vrh",
            "G_Reuss": "g_reuss",
            "G_Voigt": "g_voigt",
            "G_VRH": "g_vrh",
            "unstable": "elastically_unstable",
            "elastic_tensor": "cij",
            "vl": "vl",
            "vt": "vt",
            "vs": "vs",
        }
        if properties is None:
            properties = list(properties_map.keys())

        elastic_dict = {}
        for property in properties:
            elastic_dict[property] = getattr(self, properties_map[property])
        return elastic_dict

    @property
    def results(self) -> dict:
        if getattr(self, "_results", None) is None:
            self._results = self._analyze_elastic()
        return self._results

    @classmethod
    def from_calc_dir(
        cls,
        calc_dir: WorkingDirectory,
        **kwargs,
    ) -> ElasticAnalyzer:
        """
        Method to construct ElasticAnalyzer from a calculation directory.
        The directory should contain a POSCAR and an OUTCAR with elastic constants

        Args:
            calc_dir:

        Returns:
            Instance of ElasticAnalyzer
        """
        calc_dir = Path(calc_dir)
        if not calc_dir.exists():
            raise ValueError(f"Could not set calc_dir to {calc_dir} as it does not exist")

        structure = Structure.from_file(calc_dir / "POSCAR")
        outcar_glob = list(calc_dir.glob("OUTCAR*"))
        if len(outcar_glob) == 0:
            raise FileNotFoundError(f"No OUTCAR available in {calc_dir}")
        outcar_filepath = outcar_glob[0]

        elastic_filepath = calc_dir / "elastic_constants.txt"
        if not elastic_filepath.exists():
            # Scrapes OUTCAR for elastic constants and write to elastic_constants.txt
            elastic_table = pgrep(
                outcar_filepath,
                str_to_grep="TOTAL ELASTIC MOD",
                stop_after_first_match=True,
                after=8,
                as_string=True,
            )
            assert isinstance(elastic_table, str)
            with open(elastic_filepath, "w+") as fw:
                fw.write(elastic_table)

        # Read the elastic constants and construct the stiffness tensor
        with open(elastic_filepath) as fr:
            raw_elastic_data = fr.readlines()
        try:
            # Skip first 3 rows as they are just header
            # Skip the first column as it contains xx, xy, etc
            elastic_data = [line.strip().split()[1:] for line in raw_elastic_data[3:]]
            # Divide by 10 to get GPa instead of kBar
            elastic_tensor = np.array(elastic_data, dtype=float) / 10.0
            # VASP only outputs the stiffness tensor to 4 places
            elastic_tensor = np.round(elastic_tensor, 4)
            elastic_tensor = ElasticAnalyzer.change_elastic_constants_from_vasp(
                elastic_tensor
            )
        except Exception as e:
            raise RuntimeError(
                "Could not construct elastic tensor from OUTCAR. Please verify the"
                " OUTCAR contains a full elastic tensor.\n"
                f"{e}"
            )
        return cls(cij=elastic_tensor, structure=structure, **kwargs)



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

from __future__ import annotations

import gzip
import json
import logging
import os
from collections import deque
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from pymatgen.core import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

if TYPE_CHECKING:
    from vasp_manager.types import Filepath


@contextmanager
def change_directory(new_dir: str | Path):
    prev_dir = os.getcwd()
    os.chdir(os.path.expanduser(new_dir))
    try:
        yield
    finally:
        os.chdir(prev_dir)


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class LoggerAdapter(logging.LoggerAdapter):
    """Logging adapter to add a custom prefix to a Logger"""

    def __init__(
        self,
        logger: logging.Logger,
        prefix: str,
        separator: str = " -- ",
    ):
        super(LoggerAdapter, self).__init__(logger, {})
        self.prefix = prefix
        self.separator = separator

    def process(self, msg, kwargs):
        return f"{self.prefix}{self.separator}{msg}", kwargs


def get_pmg_structure_from_poscar(
    poscar_path: Filepath,
    to_process: bool = True,
    primitive: bool = True,
    symprec: float = 1e-03,
    angle_tolerance: float = -1.0,
    international_monoclinic: bool = False,
    return_spacegroup: bool = False,
) -> Structure | tuple[Structure, int]:
    """
    Args:
        poscar_path: path to POSCAR file
        to_process: if True, get standard reduced structure
        primitive: if True, get primitive structure, else get
            conventional structure
        symprec: symprec for SpacegroupAnalyzer
        angle_tolerance: angle tolerance for SpacegroupAnalyzer
        international_monoclinic: if True, convert to proper
            international convention such that beta is the non-right angle.
            WARNING: setting True is not compatible with pymatgen kpaths
        return_spacegroup: if True, return spacegroup number

    Returns:
        structure: structure from POSCAR
    """
    structure = Structure.from_file(poscar_path)
    if to_process:
        sga = SpacegroupAnalyzer(
            structure, symprec=symprec, angle_tolerance=angle_tolerance
        )
        if primitive:
            structure = sga.get_primitive_standard_structure(
                international_monoclinic=international_monoclinic
            )
        else:
            structure = sga.get_conventional_standard_structure(
                international_monoclinic=international_monoclinic
            )
        if return_spacegroup:
            spacegroup = sga.get_space_group_number()
            return structure, spacegroup
    return structure


def pcat(file_names: Filepath | list[Filepath]) -> str:
    """
    Custom python-only replacement for cat

    Args:
        file_names: names of files to cat together

    Returns:
        catted:
    """
    file_contents = []
    if isinstance(file_names, (str, Path)):
        file_names = [file_names]
    for file_name in file_names:
        with open(file_name) as fr:
            file_content = fr.read()
        file_contents.append(file_content)
    catted = "".join(file_content for file_content in file_contents)
    return catted


def pgrep(
    file_name: Filepath,
    str_to_grep: str,
    stop_after_first_match: bool = False,
    after: int | None = None,
    as_string: bool = False,
) -> str | list[str]:
    """
    Custom python-only replacement for grep

    Args:
        file_name: path of file
        str_to_grep: target string
        stop_after_first_match: if True, stop after first found instance of
            str_to_grep
        after: if not None, return {after} lines found after str_to_grep
        as_string: if as_string, return a single string, else return splitlines

    Returns:
        matches:
    """
    opener = gzip.open if ".gz" in str(file_name) else open
    matches = []
    line_idx_to_include: set[int] = set()
    with opener(file_name, "rt") as fr:
        for line_idx, line in enumerate(fr):
            if str_to_grep in line:
                matches.append(line.strip("\n"))
                if after is not None:
                    line_idx_to_include.update(range(line_idx + 1, line_idx + after + 1))
            if line_idx in line_idx_to_include:
                matches.append(line.strip("\n"))
                line_idx_to_include.discard(line_idx)
            if stop_after_first_match:
                if len(matches) != 0 and len(line_idx_to_include) == 0:
                    break
    if as_string:
        matches_as_str = "\n".join([line for line in matches])
    return matches_as_str if as_string else matches


def phead(
    file_name: Filepath,
    n_head: int = 1,
    as_string: bool = False,
) -> str | list[str]:
    """
    Custom python-only replacement for head

    Args:
        file_name: path of file
        n_head: n lines to head
        as_string: if as_string, return a single string, else return splitlines

    Returns:
        head:
    """
    opener = gzip.open if ".gz" in str(file_name) else open
    head = []
    with opener(file_name, "rt") as fr:
        for i, line in enumerate(fr):
            if i < n_head:
                head.append(line.strip("\n"))
    if as_string:
        head_as_str = "\n".join([line for line in head])
    return head_as_str if as_string else head


def ptail(
    file_name: Filepath,
    n_tail: int = 1,
    as_string: bool = False,
) -> str | list[str]:
    """
    Custom python-only replacement for grep

    Args:
        file_name: path of file
        n_tail: n lines to tail
        as_string: if as_string, return a single string, else return splitlines

    Returns:
        tail:
    """
    opener = gzip.open if ".gz" in str(file_name) else open
    tail: deque[str] = deque(maxlen=n_tail)
    with opener(file_name, "rt") as fr:
        for line in fr:
            tail.append(line.strip("\n"))
    if as_string:
        tail_as_str = "\n".join([line for line in tail])
    return tail_as_str if as_string else list(tail)


def make_potcar_anonymous(
    input_file_name: Filepath,
    output_file_name: Filepath | None = None,
) -> None:
    """
    Replace full POTCAR with only single POTCAR names

    Args:
        input_file_name: path of POTCAR file
        output_file_name: path to write anonymized POTCAR. If None, write to
            the location of input_f_name
    """
    if output_file_name is None:
        output_file_name = input_file_name

    with open(input_file_name, "rt") as fr:
        full_potcar_text = [line.strip() for line in fr.readlines()]
    trimmed_potcar_lines = []
    for line in full_potcar_text:
        if "TITEL" in line:
            trimmed_potcar_lines.append(line.split("=")[1].strip())
    trimmed_potcar_string = "\n".join([line for line in trimmed_potcar_lines])
    with open(output_file_name, "w+") as fw:
        fw.write(trimmed_potcar_string)



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

from __future__ import annotations

import logging
import os
import shutil
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import NDArray

from vasp_manager.analyzer.bulkmod_analyzer import BulkmodAnalyzer
from vasp_manager.calculation_manager.base import BaseCalculationManager
from vasp_manager.utils import LoggerAdapter, change_directory, pgrep
from vasp_manager.vasp_input_creator import VaspInputCreator

if TYPE_CHECKING:
    from vasp_manager.types import CalculationType, WorkingDirectory

logger = logging.getLogger(__name__)


class BulkmodCalculationManager(BaseCalculationManager):
    """
    Runs bulk modulus job workflow for a single material
    """

    def __init__(
        self,
        material_dir: WorkingDirectory,
        to_rerun: bool,
        to_submit: bool,
        primitive: bool = True,
        ignore_personal_errors: bool = True,
        from_scratch: bool = False,
        from_relax: bool = True,
        tail: int = 5,
        strains: None | NDArray = None,
    ):
        """
        Args:
            material_dir: path to a directory for a single material
            to_rerun: if True, rerun failed calculations
            to_submit: if True, submit calculations to job manager
            primitive: if True, find primitive cell, else find conventional cell
            ignore_personal_errors: if True, ignore job submission errors
                if on personal computer
            from_scratch: if True, remove the calculation's directory and
                restart
            from_relax: if True, use CONTCAR from relax
            tail: number of last lines to log in debugging if job failed
            strains: fractional strain along each axis for each deformation
                if None, use np.linspace(start=0.925, stop=1.075, number=11)**(1/3)
                len(strains) must be odd and strains must be centered around 0
        """
        self.from_relax = from_relax
        self.strains = (
            strains
            if strains is not None
            else np.power(np.linspace(0.925, 1.075, 11), 1 / 3)
        )
        self.tail = tail
        super().__init__(
            material_dir=material_dir,
            to_rerun=to_rerun,
            to_submit=to_submit,
            primitive=primitive,
            ignore_personal_errors=ignore_personal_errors,
            from_scratch=from_scratch,
        )
        self._is_done: bool
        self._results: None | str | dict
        self.logger = LoggerAdapter(logging.getLogger(__name__), self.material_name)

    @cached_property
    def mode(self) -> CalculationType:
        return "bulkmod"

    @cached_property
    def poscar_source_path(self) -> Path:
        if self.from_relax:
            poscar_source_path = self.material_dir / "rlx" / "CONTCAR"
        else:
            poscar_source_path = self.material_dir / "POSCAR"
        return poscar_source_path

    @cached_property
    def vasp_input_creator(self) -> VaspInputCreator:
        return VaspInputCreator(
            self.calc_dir,
            mode=self.mode,
            poscar_source_path=self.poscar_source_path,
            primitive=self.primitive,
            name=self.material_name,
        )

    @property
    def strains(self) -> NDArray:
        return self._strains

    @strains.setter
    def strains(self, values: NDArray) -> None:
        if np.any(values < 0.8) or np.any(values > 1.2):
            raise ValueError("Strains not in expected bounds")
        if (middle := values[int(len(values) / 2)]) != 1.0:
            raise ValueError(f"Strains not centered around 1.0: middle is {middle}")
        self._strains = values

    def _check_use_spin(self) -> bool:
        if self.from_relax:
            rlx_stdout = self.material_dir / "rlx" / "stdout.txt"
            rlx_mags = pgrep(rlx_stdout, "mag=", stop_after_first_match=True)
            use_spin = len(rlx_mags) != 0
        else:
            use_spin = True
        return use_spin

    def setup_calc(
        self,
        increase_nodes_by_factor: int = 1,
        increase_walltime_by_factor: int = 1,
    ) -> None:
        """
        Sets up an EOS bulkmod calculation
        """
        if not self.from_relax:
            msg = (
                "Running bulk modulus calculation without previous relaxation\n"
                "\tStarting structure must be fairly close to equilibrium volume!"
            )
            self.logger.warning(msg)

        self.vasp_input_creator.increase_nodes_by_factor = increase_nodes_by_factor
        self.vasp_input_creator.increase_walltime_by_factor = increase_walltime_by_factor
        self.vasp_input_creator.use_spin = self._check_use_spin()
        self.vasp_input_creator.create()
        self._make_bulkmod_strains()

        if self.to_submit:
            job_submitted = self.submit_job()
            # job status returns True if sucessfully submitted, else False
            if not job_submitted:
                self.setup_calc()

    def check_calc(self) -> bool:
        """
        Checks result of bulk modulus calculation

        Returns:
            bulkmod_sucessful: if True, bulkmod calculation completed successfully
        """
        if not self.job_complete:
            self.logger.info(f"{self.mode.upper()} job not finished")
            return False

        for i, strain in enumerate(self.strains):
            middle = int(len(self.strains) / 2)
            strain_index = i - middle
            strain_name = f"strain_{strain_index}"
            strain_dir = self.calc_dir / strain_name
            stdout_path = strain_dir / "stdout.txt"
            stderr_path = strain_dir / "stderr.txt"
            if not stdout_path.exists():
                return False

            vasp_errors = self._check_vasp_errors(
                stdout_path=stdout_path, stderr_path=stderr_path, extra_errors=["NELM"]
            )
            if len(vasp_errors) > 0:
                all_errors_addressed = self._address_vasp_errors(vasp_errors)
                if all_errors_addressed:
                    if self.to_rerun:
                        self.logger.info(f"Rerunning {self.calc_dir}")
                        self._from_scratch()
                        self.setup_calc()
                else:
                    msg = (
                        f"{self.mode.upper()} Calculation: "
                        "Couldn't address all VASP Errors\n"
                        f"\tVASP Errors: {vasp_errors}\n"
                        "\tRefusing to continue...\n"
                    )
                    self.logger.error(msg)
                    self.stop()
                return False

            grep_output = pgrep(stdout_path, "1 F=", stop_after_first_match=True)
            if len(grep_output) == 0:
                if self.to_rerun:
                    self.logger.info(f"Rerunning {self.calc_dir}")
                    # increase nodes as its likely the calculation failed
                    self._from_scratch()
                    self.setup_calc(increase_walltime_by_factor=2)
                return False
        return True

    @property
    def is_done(self) -> bool:
        if getattr(self, "_is_done", None) is None:
            self._is_done = self.check_calc()
        return self._is_done

    @property
    def results(self) -> None | str | dict:
        if not self.is_done:
            if self.stopped:
                return "STOPPED"
            else:
                return None
        try:
            ba = BulkmodAnalyzer(calc_dir=self.calc_dir)
            self._results = ba.results
            self.logger.info(f"{self.mode.upper()} Calculation: Success")
            self.logger.info(f"BULK MODULUS: {ba.results.get('B')}")
        except Exception as e:
            self.logger.warning(e)
            self._results = None
        return self._results

    def _make_bulkmod_strains(self) -> None:
        """
        Creates a set of strain directories for fitting the E-V info
        """
        self.logger.info("Making strain directories")
        for i, strain in enumerate(self.strains):
            middle = int(len(self.strains) / 2)
            strain_index = i - middle
            strain_name = f"strain_{strain_index}"
            strain_dir = self.calc_dir / strain_name
            self.logger.info(strain_dir)

            if not strain_dir.exists():
                strain_dir.mkdir()
            orig_poscar_path = self.calc_dir / "POSCAR"
            strain_poscar_path = strain_dir / "POSCAR"
            shutil.copy(orig_poscar_path, strain_poscar_path)

            # change second line to be {strain} rather than 1.0
            with open(strain_poscar_path, "r") as fr:
                strain_poscar = fr.read().splitlines()
            strain_poscar[1] = f"{strain}"
            self.logger.debug(f"{strain}")
            with open(strain_poscar_path, "w+") as fw:
                fw.write("\n".join(strain_poscar))

            with change_directory(strain_dir):
                for f in ["POTCAR", "INCAR"]:
                    if Path(f).exists():
                        os.remove(f)
                    orig_path = Path("..") / f
                    os.symlink(orig_path, f, target_is_directory=False)



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

from __future__ import annotations

import json
import logging
import os
import shutil
import stat
import warnings
from datetime import time, timedelta
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

import importlib_resources
import numpy as np
import yaml
from pymatgen.core import Structure
from pymatgen.io.vasp import Poscar, Potcar

from vasp_manager.utils import (
    LoggerAdapter,
    change_directory,
    get_pmg_structure_from_poscar,
    pcat,
)

if TYPE_CHECKING:
    from vasp_manager.types import Filepath, SourceDirectory, WorkingDirectory

logger = logging.getLogger(__name__)


class VaspInputCreator:
    """
    Handles VASP file creation
    """

    def __init__(
        self,
        calc_dir: WorkingDirectory,
        mode: str,
        poscar_source_path: Filepath,
        config_dir: SourceDirectory | None = None,
        primitive: bool = True,
        name: str = None,
        increase_nodes_by_factor: int = 1,
        increase_walltime_by_factor: int = 1,
        poscar_significant_figures: int = 16,
        ncore_per_node_for_memory: int | None = None,
        use_spin: bool = True,
    ) -> None:
        """
        Args:
            calc_dir:
            mode:
            poscar_source_path:
            config_dir:
            primitive:
            name:
            increase_nodes_by_factor:
            increase_walltime_by_factor:
            poscar_significant_figures:
            ncore_per_node_for_memory:
            use_spin: pass False to suppress spin polarization
        """
        self.calc_dir = Path(calc_dir)
        self.mode = mode
        self.poscar_source_path = Path(poscar_source_path)
        self.config_dir = Path(config_dir) if config_dir else self.calc_dir.parents[1]
        self.primitive = primitive
        self.increase_nodes_by_factor = int(increase_nodes_by_factor)
        self.increase_walltime_by_factor = int(increase_walltime_by_factor)
        self.name = name if name else self.source_structure.composition.reduced_formula
        self.poscar_significant_figures = poscar_significant_figures
        self.ncore_per_node_for_memory = ncore_per_node_for_memory
        self.use_spin = use_spin
        self.logger = LoggerAdapter(logging.getLogger(__name__), self.name)

    @cached_property
    def calc_config(self) -> dict:
        fname = "calc_config.json"
        fpath = self.config_dir / fname
        if fpath.exists():
            with open(fpath) as fr:
                calc_config_dict = json.load(fr)
                calc_config = calc_config_dict[self.mode]
        else:
            raise Exception(f"No {fname} found in path {self.config_dir.absolute()}")
        # if material_name/mode calc_config.json exists, update with that
        # This allows for custom settings for a specific material/mode to be used
        custom_calc_config_path = self.calc_dir / fname
        if custom_calc_config_path.exists():
            with open(custom_calc_config_path) as fr:
                custom_calc_config = json.load(fr)
            calc_config.update(custom_calc_config)
        return calc_config

    @cached_property
    def computing_config_dict(self) -> dict:
        """
        Dict containing all computing configs
        """
        fname = "computing_config.json"
        fpath = self.config_dir / fname
        if fpath.exists():
            with open(fpath) as fr:
                computing_config_dict = json.load(fr)
        else:
            raise Exception(f"No {fname} found in path {self.config_dir.absolute()}")
        return computing_config_dict

    @cached_property
    def computing_config(self) -> dict:
        """
        Dict containing only the computing config for the specified computer
        """
        return self.computing_config_dict[self.computer]

    @cached_property
    def computer(self) -> str:
        return self.computing_config_dict["computer"]

    @cached_property
    def source_structure(self) -> Structure:
        num_archives = len(list(self.calc_dir.glob("archive*")))
        if num_archives > 0:
            archive_name = f"archive_{num_archives-1}"
            self.poscar_source_path = self.calc_dir / archive_name / "CONTCAR"
        try:
            structure = get_pmg_structure_from_poscar(
                self.poscar_source_path, primitive=self.primitive
            )
        except Exception as e:
            raise Exception(f"Cannot load POSCAR in {self.poscar_source_path}: {e}")
        assert isinstance(structure, Structure)
        return structure

    @cached_property
    def incar_template(self) -> str:
        incar_template = (
            importlib_resources.files("vasp_manager")
            .joinpath(str(Path("static_files") / "INCAR_template"))
            .read_text()
        )
        return incar_template

    @cached_property
    def potcar_dict(self) -> dict:
        potcar_dict = json.loads(
            importlib_resources.files("vasp_manager")
            .joinpath(str(Path("static_files") / "pot_dict.json"))
            .read_text()
        )
        return potcar_dict

    @cached_property
    def q_mapper(self) -> dict:
        q_mapper = json.loads(
            importlib_resources.files("vasp_manager")
            .joinpath(str(Path("static_files") / "q_handles.json"))
            .read_text()
        )
        return q_mapper

    def make_poscar(self) -> None:
        """
        Create and write a POSCAR
        """
        poscar = Poscar(self.source_structure)
        poscar_path = self.calc_dir / "POSCAR"
        poscar.write_file(
            poscar_path, significant_figures=self.poscar_significant_figures
        )

    def make_potcar(self) -> None:
        """
        Create and write a POTCAR
        """
        potcar_path = self.calc_dir / "POTCAR"
        potcar_dir = Path(self.computing_config["potcar_dir"])

        el_names = [el.name for el in self.source_structure.composition]
        self.logger.debug(
            f"{self.source_structure.composition.reduced_formula}, {el_names}"
        )
        pot_singles = [
            potcar_dir / self.potcar_dict[el_name] / "POTCAR" for el_name in el_names
        ]
        for pot_single in pot_singles:
            if not pot_single.exists():
                msg = f"Unable to create POTCAR\n\tPOTCAR not found at path {pot_single}"
                raise Exception(msg)

        potcar = pcat(pot_singles)
        with open(potcar_path, "w+") as fw:
            fw.write(potcar)

    @cached_property
    def n_nodes(self) -> int:
        # start with 1 node per 32 atoms
        num_nodes = (len(self.source_structure) // 32) + 1
        if self.computer == "quest":
            # quest has ~2x smaller nodes than perlmutter
            num_nodes *= 2
        num_nodes *= self.increase_nodes_by_factor
        return num_nodes

    @cached_property
    def n_procs(self) -> int:
        n_procs = self.n_nodes * self.computing_config["ncore_per_node"]
        return n_procs

    @property
    def ncore_per_node_for_memory(self) -> int:
        return self._ncore_per_node_for_memory

    @ncore_per_node_for_memory.setter
    def ncore_per_node_for_memory(self, value: int | None) -> None:
        if value is None:
            value = 4 if self.computer == "quest" else 0
        if not isinstance(value, int):
            raise TypeError("ncore_per_node_for_memory must be an int")
        self._ncore_per_node_for_memory = value

    @cached_property
    def n_procs_used(self) -> int:
        """
        Total number of actual processors to run VASP, which may be different
        than the total number of processors requested in SLURM
            Typically request all available processors on each node, and then
            (optionally) leave some empty for memory or ease of division
        """
        ncore_per_node = self.computing_config["ncore_per_node"]
        return self.n_nodes * (ncore_per_node - self.ncore_per_node_for_memory)

    @cached_property
    def d_f_block(self) -> dict:
        d_f_block = json.loads(
            importlib_resources.files("vasp_manager")
            .joinpath(str(Path("static_files") / "d_f_block.json"))
            .read_text()
        )
        return d_f_block

    def _check_needs_spin_polarization(self, composition_dict: dict) -> bool:
        needs_spin_polarization = False
        for el_name in composition_dict:
            if (
                el_name in self.d_f_block["d_block"]
                or el_name in self.d_f_block["f_block"]
            ):
                needs_spin_polarization = True
        return needs_spin_polarization

    def _get_auto_magmom(self, composition_dict: dict) -> str:
        magmom_string = "MAGMOM ="
        for el_name in composition_dict:
            if el_name in self.d_f_block["d_block"]:
                init_mag = "5.0"
            elif el_name in self.d_f_block["f_block"]:
                init_mag = "7.0"
            else:
                init_mag = "0.0"
            magmom_atom = str(int(composition_dict[el_name])) + "*" + init_mag
            magmom_string += f" {magmom_atom}"
        return magmom_string

    @cached_property
    def hubbards(self) -> dict:
        hubbards = json.loads(
            importlib_resources.files("vasp_manager")
            .joinpath(str(Path("static_files") / "hubbards.json"))
            .read_text()
        )
        return hubbards

    def _check_needs_dftu(self, hubbards_type: str, composition_dict: dict) -> bool:
        if not hubbards_type:
            return False

        needs_dftu = False
        for el_name in composition_dict:
            if el_name in self.hubbards[hubbards_type]:
                needs_dftu = True
        return needs_dftu

    def _get_ldau_string(self, hubbards_type: str, composition_dict: dict) -> str:
        u_string = "LDAUU ="
        j_string = "LDAUJ ="
        l_string = "LDAUL ="
        hubbards = self.hubbards[hubbards_type]

        for el_name in composition_dict:
            if el_name in hubbards:
                u_val = hubbards[el_name]["U"]
                l_val = hubbards[el_name]["L"]
            else:
                u_val = 0.0
                l_val = -1
            u_string += f" {float(u_val)}"
            j_string += " 0.0"
            l_string += f" {l_val}"

        ldau_string = "LDAU = .TRUE.\n"
        ldau_string += "LDAUPRINT = 1\n"
        ldau_string += f"{u_string}\n"
        ldau_string += f"{j_string}\n"
        ldau_string += f"{l_string}"
        return ldau_string

    def _get_lmaxmix(self, composition_dict: dict) -> int:
        d_block_elements = self.d_f_block["d_block"]
        f_block_elements = self.d_f_block["f_block"]
        lmaxmix = 2
        for element in composition_dict:
            if element in d_block_elements:
                lmaxmix = 4
        for element in composition_dict:
            if element in f_block_elements:
                lmaxmix = 6
        return lmaxmix

    def make_incar(self) -> None:
        """
        Create and write an INCAR

        Current kpoints coming from the kspacing tag in the INCAR,
            but future versions should include ability to make kpoints from kppra
        """
        incar_path = self.calc_dir / "INCAR"
        calc_config = self.calc_config.copy()

        if calc_config["ispin"] not in [1, "auto"]:
            raise RuntimeError("ISPIN must be set to 1 or auto")

        if calc_config["hubbards"] not in [None, False, "wang"]:
            raise RuntimeError("hubbards must be set to None, False, or 'wang'")
        if calc_config["hubbards"] == "wang" and calc_config["gga"] != "PE":
            raise RuntimeError("DFT+U is only support for PBE")

        if calc_config["iopt"] != 0 and calc_config["potim"] != 0:
            raise RuntimeError("To use IOPT != 0, POTIM must be set to 0")

        composition_dict = self.source_structure.composition.as_dict()
        # read POTCAR
        potcar_path = self.calc_dir / "POTCAR"
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            potcar = Potcar.from_file(potcar_path)
        n_electrons = 0
        for potcar_single in potcar:
            n_electrons += (
                potcar_single.nelectrons * composition_dict[potcar_single.element]
            )
        # make n_bands divisible by NCORE (VASP INCAR tag)
        # elastic calculation won't run unless NCORE=1
        if self.mode == "elastic":
            self.computing_config["ncore"] = 1
        ncore = self.computing_config["ncore"]
        calc_config["ncore"] = ncore
        cores_per_group = int(self.n_procs_used / self.calc_config["kpar"])
        calc_config["nbands"] = np.max(
            [
                int(np.ceil(0.75 * n_electrons / ncore) * ncore),
                cores_per_group,
            ]
        )
        for write_tag in ["lcharge", "lwave", "lvtot"]:
            calc_config[write_tag] = calc_config.get(write_tag, ".FALSE.")
        calc_config["amix"] = calc_config.get("amix", 0.4)
        calc_config["bmix"] = calc_config.get("bmix", 1.0)

        needs_spin_polarization = self._check_needs_spin_polarization(composition_dict)
        use_spin_polarization = (
            needs_spin_polarization and calc_config["ispin"] == "auto" and self.use_spin
        )
        if use_spin_polarization:
            ispin = 2
            magmom_line = self._get_auto_magmom(composition_dict)
        else:
            ispin = 1

        needs_dftu = self._check_needs_dftu(calc_config["hubbards"], composition_dict)
        use_dftu = needs_dftu and calc_config["hubbards"]
        if use_dftu:
            lmaxmix = self._get_lmaxmix(composition_dict)
            lmaxmix_line = f"LMAXMIX = {lmaxmix}"
            ldau_string = self._get_ldau_string(calc_config["hubbards"], composition_dict)

        # Add lines to the incar file
        incar_tmp = self.incar_template.split("\n")
        for i, line in enumerate(incar_tmp):
            # add extra flags for spin polarization
            if "ISPIN" in line:
                incar_tmp[i] = f"ISPIN = {ispin}"
                if use_spin_polarization:
                    incar_tmp.insert(i + 1, magmom_line)
            if "LVTOT" in line and use_spin_polarization:
                lorbit_line = "LORBIT = 11"
                incar_tmp.insert(i + 1, lorbit_line)
            # add extra flags for DFT+U
            if "DFT+U" in line and use_dftu:
                incar_tmp.insert(i + 1, lmaxmix_line)
                incar_tmp.insert(i + 1, ldau_string)
            # add extra flags for elastic mode
            if self.mode == "elastic":
                if "KSPACING" in line:
                    nfree_line = "NFREE = {nfree}"
                    incar_tmp.insert(i + 1, nfree_line)
        incar_tmp_as_str = "\n".join([line for line in incar_tmp])
        incar = incar_tmp_as_str.format(**calc_config)
        self.logger.debug(incar)
        with open(incar_path, "w+") as fw:
            fw.write(incar)

    def make_kpoints(self) -> None:
        kpoints_path = self.calc_dir / "KPOINTS"
        kspacing = self.calc_config["kspacing"]
        reciprocal_lattice = self.source_structure.lattice.reciprocal_lattice
        abc = np.asarray(reciprocal_lattice.abc)
        kpoints = [int(k) for k in np.ceil(abc / kspacing)]
        kpoints_str = " ".join(str(kpoint) for kpoint in kpoints)
        kpoints_text = f"kpoints generated from kspacing={kspacing}\n"
        kpoints_text += "0\n"
        kpoints_text += "Gamma\n"
        kpoints_text += f"{kpoints_str}\n"
        with open(kpoints_path, "w+") as fw:
            fw.write(kpoints_text)

    def make_vaspq(self) -> None:
        """
        Create and write vasp.q file
        """
        vaspq_path = self.calc_dir / "vasp.q"

        # create pad string for job naming to differentiate in the queue
        match self.mode:
            case "rlx-coarse" | "rlx":
                if self.mode == "rlx":
                    pad_string = "r"
                elif self.mode == "rlx-coarse":
                    pad_string = "rc"
                mode = "rlx"
            case "static":
                pad_string = "s"
                mode = "static"
            case "bulkmod":
                pad_string = "b"
                mode = "bulkmod"
            case "elastic":
                pad_string = "e"
                mode = "elastic"
            case _:
                raise ValueError(
                    "Calculation type {self.mode} not in supported calculation types"
                    "of VaspInputCreator"
                )

        jobname = pad_string + self.name
        # convert walltime into seconds for increase_walltime_by_factor
        walltime_iso = time.fromisoformat(self.calc_config["walltime"])
        # walltime_duration is in seconds
        walltime_duration = timedelta(
            hours=walltime_iso.hour,
            minutes=walltime_iso.minute,
            seconds=walltime_iso.second,
        )
        walltime_duration *= self.increase_walltime_by_factor
        # convert to HH:MM:SS
        walltime = str(walltime_duration)
        # cut walltime short by 1 minute so job metrics log properly
        timeout_as_delta = timedelta(seconds=walltime_duration.seconds - 60)
        if self.computer == "quest":
            # quest uses mpirun which needs timeout in seconds
            timeout = str(timeout_as_delta.seconds)
        else:
            # otherwise, convert it back to HH:MM:SS
            timeout = str(timeout_as_delta)

        computing_config = self.computing_config.copy()
        ncore_per_node = self.n_procs_used // self.n_nodes
        computing_config.update(
            {
                "n_nodes": self.n_nodes,
                "n_procs": self.n_procs,
                "ncore_per_node": ncore_per_node,
                "jobname": jobname,
                "walltime": walltime,
                "timeout": timeout,
            }
        )

        vaspq_settings_path = self.q_mapper[self.computer][mode]
        vaspq_settings = yaml.load(
            importlib_resources.files("vasp_manager")
            .joinpath(str(Path("static_files") / vaspq_settings_path))
            .read_text(),
            Loader=yaml.SafeLoader,
        )
        override_vaspq_settings_path = self.config_dir / f"{self.computer}.yml"
        if override_vaspq_settings_path.exists():
            with open(override_vaspq_settings_path) as fr:
                override_vaspq_settings = yaml.load(
                    fr,
                    Loader=yaml.SafeLoader,
                )
            vaspq_settings.update(override_vaspq_settings)
        vaspq_tmp = (
            importlib_resources.files("vasp_manager")
            .joinpath(str(Path("static_files") / "vasp.q"))
            .read_text()
        )
        vaspq_tmp = vaspq_tmp.format(**vaspq_settings)
        vaspq = vaspq_tmp.format(**computing_config)
        self.logger.debug(vaspq)
        with open(vaspq_path, "w+") as fw:
            fw.write(vaspq)
        vaspq_path.chmod(vaspq_path.stat().st_mode | stat.S_IEXEC)

    def make_archive_and_repopulate(self) -> None:
        """
        Make an archive of a VASP calculation and copy back over relevant files
        """
        with change_directory(self.calc_dir):
            contcar_path = Path("CONTCAR")
            contcar_exists = contcar_path.exists()
            if contcar_exists:
                contcar_is_empty = contcar_path.stat().st_size == 0
            else:
                contcar_is_empty = True

            # if CONTCAR is empty, don't make an archive and clean up
            if contcar_is_empty:
                all_files = [
                    f
                    for f in Path(".").glob("*")
                    if f.is_file() and "archive" not in f.name and "json" not in f.name
                ]
                for f in all_files:
                    os.remove(f)
            # else, make the archive
            else:
                num_previous_archives = len(list(Path(".").glob("archive*")))
                archive_name = Path(f"archive_{num_previous_archives}")
                self.logger.info(f"Making {archive_name}...")
                archive_name.mkdir()

                all_files = [
                    f
                    for f in Path(".").glob("*")
                    if f.is_file() and "archive" not in f.name and "json" not in f.name
                ]
                for f in all_files:
                    # add if symlink for testing
                    if f.is_symlink():
                        f_links_to = f.readlink()
                        os.remove(f)
                        shutil.copy2(f_links_to, f)
                    shutil.move(f, archive_name)

        self.create()

    def create(self) -> None:
        """
        Make VASP input files

        Note:
            Don't touch the order! make_incar and make_vaspq rely on the poscar and
            potcar existing already
        """
        if not self.calc_dir.exists():
            self.calc_dir.mkdir()
        self.make_poscar()
        self.make_potcar()
        if "write_kpoints" in self.calc_config:
            if self.calc_config["write_kpoints"]:
                self.make_kpoints()
        self.make_incar()
        self.make_vaspq()



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

from __future__ import annotations

import logging
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np

from vasp_manager.calculation_manager.base import BaseCalculationManager
from vasp_manager.utils import LoggerAdapter, get_pmg_structure_from_poscar, pgrep, ptail
from vasp_manager.vasp_input_creator import VaspInputCreator

if TYPE_CHECKING:
    from vasp_manager.types import CalculationType, WorkingDirectory

logger = logging.getLogger(__name__)


class RlxCalculationManager(BaseCalculationManager):
    """
    Runs relaxation job workflow for a single material
    """

    def __init__(
        self,
        material_dir: WorkingDirectory,
        to_rerun: bool,
        to_submit: bool,
        primitive: bool = True,
        ignore_personal_errors: bool = True,
        from_scratch: bool = False,
        from_coarse_relax: bool = True,
        tail: int = 5,
        max_reruns: int = 3,
        magmom_per_atom_cutoff: float = 0.0,
    ):
        """
        Args:
            material_dir: path to a directory for a single material
            to_rerun: if True, rerun failed calculations
            to_submit: if True, submit calculations to job manager
            primitive: if True, find primitive cell, else find conventional cell
            ignore_personal_errors: if True, ignore job submission errors
                if on personal computer
            from_scratch: if True, remove the calculation's directory and
                restart
            from_coarse_relax: if True, use CONTCAR from coarse relax
            tail: number of last lines to log in debugging if job failed
            max_reruns: maximum number of times to rerun rlx before refusing to
                continue
            magmom_per_atom_cutoff: calculations that result in
                magmom_per_atom less than this parameter will be automatically
                rerun without spin-polarization
        """
        self.from_coarse_relax = from_coarse_relax
        self.tail = tail
        self.max_reruns = max_reruns
        self.magmom_per_atom_cutoff = magmom_per_atom_cutoff
        super().__init__(
            material_dir=material_dir,
            to_rerun=to_rerun,
            to_submit=to_submit,
            primitive=primitive,
            ignore_personal_errors=ignore_personal_errors,
            from_scratch=from_scratch,
        )
        self._is_done: bool
        self._results: None | str | dict
        self.logger = LoggerAdapter(logging.getLogger(__name__), self.material_name)

    @cached_property
    def mode(self) -> CalculationType:
        return "rlx"

    @cached_property
    def poscar_source_path(self) -> Path:
        if self.from_coarse_relax:
            poscar_source_path = self.material_dir / "rlx-coarse" / "CONTCAR"
        else:
            poscar_source_path = self.material_dir / "POSCAR"
        return poscar_source_path

    @cached_property
    def vasp_input_creator(self) -> VaspInputCreator:
        return VaspInputCreator(
            self.calc_dir,
            mode=self.mode,
            poscar_source_path=self.poscar_source_path,
            primitive=self.primitive,
            name=self.material_name,
        )

    def setup_calc(
        self,
        increase_nodes_by_factor: int = 1,
        increase_walltime_by_factor: int = 1,
        make_archive: bool = False,
        use_spin: bool = True,
    ) -> None:
        """
        Sets up a fine relaxation
        """
        self.vasp_input_creator.increase_nodes_by_factor = increase_nodes_by_factor
        self.vasp_input_creator.increase_walltime_by_factor = increase_walltime_by_factor
        self.vasp_input_creator.use_spin = use_spin

        if make_archive:
            self.vasp_input_creator.make_archive_and_repopulate()
        else:
            self.vasp_input_creator.create()

        if self.to_submit:
            job_submitted = self.submit_job()
            if not job_submitted:
                self.setup_calc()

    def check_calc(self) -> bool:
        """
        Checks if calculation has finished and reached required accuracy

        Returns:
            relaxation_successful: if True, relaxation completed successfully
        """
        if not self.job_complete:
            self.logger.info(f"{self.mode.upper()} not finished")
            return False

        stdout_path = self.calc_dir / "stdout.txt"
        if not stdout_path.exists():
            # calculation never actually ran
            # shouldn't get here unless function was called with submit=False
            # or job was manually cancelled
            self.logger.info(f"{self.mode.upper()} Calculation: No stdout.txt available")
            if self.to_rerun:
                self._cancel_previous_job()
                self.setup_calc()
            return False

        vasp_errors = self._check_vasp_errors()
        if len(vasp_errors) > 0:
            all_errors_addressed = self._address_vasp_errors(vasp_errors)
            if all_errors_addressed:
                if self.to_rerun:
                    self.logger.info(f"Rerunning {self.calc_dir}")
                    self.setup_calc(make_archive=True)
            else:
                msg = (
                    f"{self.mode.upper()} Calculation: "
                    "Couldn't address all VASP Errors\n"
                    f"\tVASP Errors: {vasp_errors}\n"
                    "\tRefusing to continue...\n"
                )
                self.logger.error(msg)
                self.stop()
            return False

        previous_magmom_per_atom = self._parse_magmom_per_atom()
        if previous_magmom_per_atom is None:
            use_spin = False
        else:
            use_spin = np.abs(previous_magmom_per_atom) >= self.magmom_per_atom_cutoff

        tail_output = ptail(stdout_path, n_tail=self.tail, as_string=True)
        grep_output = pgrep(
            stdout_path, "reached required accuracy", stop_after_first_match=True
        )
        if len(grep_output) == 0:
            archive_dirs = list(self.calc_dir.glob("archive*"))
            if len(archive_dirs) >= self.max_reruns - 1:
                msg = (
                    "Many archives exist, calculations may not be converging\n"
                    "\tRefusing to continue..."
                )
                self.logger.error(msg)
                return False

            self.logger.warning(f"{self.mode.upper()} FAILED")
            self.logger.debug(tail_output)
            if self.to_rerun:
                self.logger.info(f"Rerunning {self.calc_dir}")
                # increase nodes as its likely the calculation failed
                self.setup_calc(
                    increase_walltime_by_factor=2, make_archive=True, use_spin=use_spin
                )
            return False

        # in the case that the calculation finishes but results in a spin value lower
        # than the cutoff, rerun it without spin
        if not use_spin and previous_magmom_per_atom is not None:
            if self.to_rerun:
                self.logger.info(f"Rerunning {self.calc_dir}")
                self.setup_calc(make_archive=True, use_spin=False)
            return False

        self.logger.info(f"{self.mode.upper()} Calculation: reached required accuracy")
        self.logger.debug(tail_output)
        return True

    def check_volume_difference(self) -> bool:
        """
        Checks relaxation runs for volume difference

        if abs(volume difference) is >= 5%, reruns relaxation
        only checks for mode='rlx' as that's the structure for further calculation

        Returns:
            volume_converged: if True, relaxation completed successfully
        """
        original_poscar_path = self.material_dir / "POSCAR"
        poscar_path = self.calc_dir / "POSCAR"
        contcar_path = self.calc_dir / "CONTCAR"
        try:
            orig_structure, orig_spacegroup = get_pmg_structure_from_poscar(
                original_poscar_path, return_spacegroup=True
            )
            p_structure, p_spacegroup = get_pmg_structure_from_poscar(
                poscar_path, return_spacegroup=True
            )
            c_structure, c_spacegroup = get_pmg_structure_from_poscar(
                contcar_path, return_spacegroup=True
            )
        except Exception as e:
            self.logger.error(f"RLX CONTCAR doesn't exist or is empty: {e}")
            return False

        if orig_spacegroup == c_spacegroup:
            self.logger.debug(
                f"Spacegroups match orig-{orig_spacegroup} == rlx-{c_spacegroup}"
            )
        else:
            self.logger.warning(
                "Warning: spacegroups do not match "
                + f"orig-{orig_spacegroup} != rlx-{c_spacegroup}"
            )

        volume_diff = (c_structure.volume - p_structure.volume) / p_structure.volume
        if np.abs(volume_diff) >= 0.05:
            self.logger.warning(f"NEED TO RE-RELAX: dV = {volume_diff:.4f}")
            volume_converged = False
            previous_magmom_per_atom = self._parse_magmom_per_atom()
            if previous_magmom_per_atom is None:
                use_spin = False
            else:
                use_spin = np.abs(previous_magmom_per_atom) >= self.magmom_per_atom_cutoff
            if self.to_rerun:
                self.setup_calc(make_archive=True, use_spin=use_spin)
        else:
            self.logger.info("RLX volume converged")
            self.logger.debug(f"dV = {volume_diff:.4f}")
            volume_converged = True
        orig_volume_diff = (
            c_structure.volume - orig_structure.volume
        ) / orig_structure.volume
        self._results = {}
        self._results["initial_spacegroup"] = orig_spacegroup
        self._results["relaxed_spacegroup"] = c_spacegroup
        self._results["total_dV"] = np.round(orig_volume_diff, 4)
        return volume_converged

    @property
    def is_done(self) -> bool:
        if getattr(self, "_is_done", None) is None:
            self._is_done = False
            calc_done = self.check_calc()
            if calc_done:
                volume_converged = self.check_volume_difference()
                if volume_converged:
                    self._is_done = True
        return self._is_done

    @property
    def results(self) -> None | str | dict:
        if not self.is_done:
            if self.stopped:
                return "STOPPED"
            else:
                return None
        return self._results



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

from __future__ import annotations

import os
import shutil
from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from vasp_manager.job_manager import JobManager
from vasp_manager.utils import get_pmg_structure_from_poscar, pgrep
from vasp_manager.vasp_input_creator import VaspInputCreator

if TYPE_CHECKING:
    from vasp_manager.types import CalculationType, WorkingDirectory


class BaseCalculationManager(ABC):
    """
    Runs vasp job workflow for a single material
    """

    def __init__(
        self,
        material_dir: WorkingDirectory,
        to_rerun: bool,
        to_submit: bool,
        primitive: bool = True,
        ignore_personal_errors: bool = True,
        from_scratch: bool = False,  # DANGEROUS, WILL DELETE PREVIOUS CALCULATION
    ):
        """
        Args:
            material_dir: path to a directory for a single material
                ex. calculations/AlAs
            to_rerun: if True, rerun failed calculations
            to_submit: if True, submit calculations to job manager
            primitive: if True, find primitive cell, else find conventional cell
            ignore_personal_errors: if True, ignore job submission errors
                if on personal computer
            from_scratch: if True, remove the calculation's directory and
                restart
                note: DANGEROUS
        """
        self.material_dir = Path(material_dir)
        self.to_rerun = to_rerun
        self.to_submit = to_submit
        self.primitive = primitive
        self.job_manager = JobManager(
            calc_dir=self.calc_dir,
            manager_name=f"{self.material_name} {self.mode.upper()}",
            ignore_personal_errors=ignore_personal_errors,
        )

        self.from_scratch = from_scratch
        if from_scratch:
            self._from_scratch()

    @property
    @abstractmethod
    def mode(self) -> CalculationType:
        pass

    @property
    @abstractmethod
    def is_done(self) -> bool:
        pass

    @property
    @abstractmethod
    def vasp_input_creator(self) -> VaspInputCreator:
        pass

    @cached_property
    def calc_dir(self) -> Path:
        return self.material_dir / self.mode

    @cached_property
    def material_name(self) -> str:
        return self.material_dir.name

    @cached_property
    @abstractmethod
    def poscar_source_path(self) -> Path:
        pass

    @abstractmethod
    def setup_calc(self) -> None:
        pass

    @abstractmethod
    def check_calc(self) -> bool:
        pass

    @property
    def job_exists(self) -> bool:
        return self.job_manager.job_exists

    @property
    def job_complete(self) -> bool:
        return self.job_manager.job_complete

    @property
    def stopped(self) -> bool:
        return (self.material_dir / "STOP").exists()

    def stop(self) -> None:
        with open(self.material_dir / "STOP", "w+"):
            pass

    def submit_job(self) -> bool:
        return self.job_manager.submit_job()

    def _cancel_previous_job(self) -> None:
        jobid_path = self.calc_dir / "jobid"
        if jobid_path.exists():
            with open(jobid_path) as fr:
                jobid = fr.read().strip()
            cancel_job_call = f"scancel {jobid}"
            os.system(cancel_job_call)
            os.remove(jobid_path)

    def _from_scratch(self) -> None:
        self._cancel_previous_job()
        shutil.rmtree(self.calc_dir)

    def _parse_magmom(self) -> float | None:
        stdout_path = self.calc_dir / "stdout.txt"
        mag_lines = pgrep(stdout_path, "mag=")
        # if "mag=" not found in stdout, set magmom=None
        if len(mag_lines) == 0:
            magmom = None
        else:
            total_mag = mag_lines[-1].split()[-1]
            magmom = float(total_mag)
        return magmom

    def _parse_magmom_per_atom(self) -> float | None:
        total_magmom = self._parse_magmom()
        if total_magmom is None:
            return None
        structure = get_pmg_structure_from_poscar(self.poscar_source_path)
        magmom_per_atom = total_magmom / len(structure)
        return magmom_per_atom

    def _parse_incar_tag(self, tag) -> str:
        incar_path = self.calc_dir / "INCAR"
        with open(incar_path) as fr:
            incar = fr.readlines()
        for line in incar:
            if tag in line:
                tag_value = line.split("=")[1].strip()
        return tag_value

    def _check_vasp_errors(
        self,
        stdout_path: Path | None = None,
        stderr_path: Path | None = None,
        extra_errors: list[str] | None = None,
    ) -> set[str]:
        """
        Find VASP errors in stdout and stderr

        Args:
            stdout_path
            stderr_path
            extra_errors: names of other errors to include in search

        Returns:
            errors_found: True if any errors found
        """
        if stdout_path is None:
            stdout_path = self.calc_dir / "stdout.txt"
        if stderr_path is None:
            stderr_path = self.calc_dir / "stderr.txt"
        if extra_errors is None:
            extra_errors = []
        errors_found = set()

        with open(stdout_path) as fr:
            for line in fr:
                errors = [
                    "Sub-Space-Matrix",
                    "Inconsistent Bravais",
                    "num prob",
                    "BRMIX",
                    "SICK JOB",
                    "VERY BAD NEWS",
                    "Fatal error",
                ]
                errors.extend(extra_errors)
                for error in errors:
                    if error in line:
                        errors_found.add(error)

        with open(stderr_path) as fr:
            for line in fr:
                for error in [
                    "oom-kill",
                    "SETYLM",
                    "Segmentation",
                    "command not found",
                ]:
                    if error in line:
                        errors_found.add(error)

        return errors_found

    def _address_vasp_errors(self, errors: set[str]) -> bool:
        """
        Args:
            errors: set of errors found in stdout or stderr

        Returns:
            all_errors_addressed: if True, all errors could be fixed
                automatically. If False, some errors could not be handled
        """
        vic = self.vasp_input_creator

        errors_addressed = {e: False for e in errors}
        for error in errors:
            match error:
                case "Sub-Space-Matrix":
                    new_algo = "Fast"
                    previous_algo = self._parse_incar_tag("ALGO")
                    if previous_algo == new_algo:
                        errors_addressed[error] = False
                    else:
                        vic.calc_config["algo"] = new_algo
                        errors_addressed[error] = True
                case "Inconsistent Bravais":
                    new_symprec = "1e-08"
                    previous_symprec = self._parse_incar_tag("SYMPREC")
                    if previous_symprec == new_symprec:
                        errors_addressed[error] = False
                    else:
                        vic.calc_config["symprec"] = new_symprec
                        errors_addressed[error] = True
                case "oom-kill":
                    if vic.computer == "quest":
                        ncore_per_node_for_memory = 28
                    else:
                        ncore_per_node_for_memory = 64
                    vic.ncore_per_node_for_memory = ncore_per_node_for_memory
                    errors_addressed[error] = True
                case _:
                    errors_addressed[error] = False
        all_errors_addressed = all([v for v in errors_addressed.values()])
        return all_errors_addressed



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Literal, TypeAlias

    import numpy as np

    from vasp_manager.calculation_manager.bulkmod import BulkmodCalculationManager
    from vasp_manager.calculation_manager.elastic import ElasticCalculationManager
    from vasp_manager.calculation_manager.rlx import RlxCalculationManager
    from vasp_manager.calculation_manager.rlx_coarse import RlxCoarseCalculationManager
    from vasp_manager.calculation_manager.static import StaticCalculationManager

    Filepath = str | Path
    Filepaths = list[Path]
    SourceDirectory = str | Path
    WorkingDirectory = str | Path
    Floating = float | np.floating
    CalculationType = Literal["rlx-coarse", "rlx", "static", "bulkmod", "elastic"]
    CalculationManager: TypeAlias = (
        BulkmodCalculationManager
        | ElasticCalculationManager
        | RlxCalculationManager
        | RlxCoarseCalculationManager
        | StaticCalculationManager
    )
